
--[[ functions used by more than one module ]]

local _,L = ...
local rematch = Rematch
local settings

rematch.info = {} -- reusable table to reduce garbage creation
rematch.abilityList = {} -- reusable table for C_PetJournal.GetPetAbilityList
rematch.levelList = {} -- reusable table for C_PetJournal.GetPetAbilityList
rematch.emptyPetID = "0x0000000000000000" -- petID to load for an empty slot

rematch.WoDBuild = select(4,GetBuildInfo())==60000

BINDING_HEADER_REMATCH = "Rematch"
BINDING_NAME_REMATCH_WINDOW = L["Toggle Window"]
BINDING_NAME_REMATCH_AUTOLOAD = L["Toggle Auto Load"]
BINDING_NAME_REMATCH_PETS = L["Toggle Pets"]
BINDING_NAME_REMATCH_TEAMS = L["Toggle Teams"]

function rematch:InitCommon()
	RematchSettings = RematchSettings or {}
	settings = RematchSettings
	settings.LevelingQueue = settings.LevelingQueue or {}
	RematchSaved = RematchSaved or {}

	if rematch.WoDBuild then
		if settings.LevelingQueue[1] and not settings.LevelingQueue[1]:match("BattlePet") then
			wipe(settings.LevelingQueue)
		end
	end

	rematch.breedable = IsAddOnLoaded("BattlePetBreedID")
	hooksecurefunc(C_PetJournal,"PickupPet",rematch.events.CURSOR_UPDATE)
	rematch:RegisterEvent("PET_JOURNAL_LIST_UPDATE")

	rematch:RegisterEvent("PLAYER_TARGET_CHANGED")
	rematch:RegisterEvent("UPDATE_MOUSEOVER_UNIT")

	rematch:RegisterEvent("PLAYER_REGEN_ENABLED")
	rematch:RegisterEvent("PLAYER_REGEN_DISABLED")
	rematch:RegisterEvent("PET_BATTLE_OPENING_START")
	rematch:RegisterEvent("PET_BATTLE_OVER")

	local bindsMoved
	for oldBind,newBind in pairs({["Toggle Rematch Window"]="WINDOW", ["Toggle Auto Load"]="AUTOLOAD", ["Toggle Pet Browser"]="PETS"}) do
		local key = GetBindingKey(oldBind)
		if key then
			SetBinding(key,"REMATCH_"..newBind)
			bindsMoved = true
		end
	end
	if bindsMoved then
		SaveBindings(GetCurrentBindingSet())
	end

end

--[[ Events ]]

rematch.events = {}
function rematch:OnEvent(event,...)
	if rematch.events[event] then
		rematch.events[event](...)
	end
end

function rematch.events.PLAYER_LOGIN()
	rematch:InitCommon()
	rematch:InitMain()
	rematch:InitCurrent()
	rematch:InitOptions()
	rematch:InitBrowser()
	rematch:InitLeveling()
	rematch:InitMenu()
	rematch:InitRMF()
	rematch:InitDialog()
	rematch:InitTeams()
	rematch:InitSaveAs()
	rematch:InitPetLoading()
end


function rematch.events.PET_JOURNAL_LIST_UPDATE()
	local _,owned = C_PetJournal.GetNumPets()
	if owned and owned~=rematch.lastOwned then -- if number of owned pets changed
		rematch.lastOwned = owned
		rematch:ProcessQueue() -- immediately process the queue
		rematch.finderNeedsUpdate = true -- and flag find pet system that its data isn't current
	end
	rematch:StartTimer("UpdateWindow",0,rematch.UpdateWindow)
end

function rematch.events.PLAYER_REGEN_DISABLED()
	if rematch:IsVisible() then
		rematch.resummonWindow = true
		rematch:Toggle()
	end
	rematch:UnregisterEvent("PLAYER_TARGET_CHANGED")
	rematch:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")
end

function rematch.events.PLAYER_REGEN_ENABLED()
	if rematch.resummonWindow then
		rematch:Show()
		rematch.resummonWindow = nil
	end
	if rematch.queueNeedsProcessed then
		rematch:ProcessQueue()
		rematch.queueNeedsProcessed = nil
	end
	if rematch.teamNeedsLoaded then
		rematch:LoadTeam(rematch.teamNeedsLoaded)
		rematch.teamNeedsLoaded = nil
	end
	rematch:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	rematch:RegisterEvent("PLAYER_TARGET_CHANGED")
end

function rematch.events.PET_BATTLE_OPENING_START()
	if not settings.StayForBattle then
		rematch.returnAfterBattle = rematch:IsVisible()
		rematch:Hide()
	end
	rematch:UnregisterEvent("PLAYER_TARGET_CHANGED")
	rematch:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")
end

function rematch.events.PET_BATTLE_OVER()
	rematch:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	rematch:RegisterEvent("PLAYER_TARGET_CHANGED")
	rematch:ProcessQueue()
	if rematch.returnAfterBattle then
		rematch:Show()
		rematch.returnAfterBattle = nil
	end
	if rematch.teamNeedsLoaded then
		rematch:LoadTeam(teamNeedsLoaded)
		rematch.teamNeedsLoaded = nil
	end
end

--[[ Pet template script handlers ]]

function rematch:PetOnEnter()
	rematch:ShowFloatingPetCard(rematch:GetPetID(self.petID),self,self.menu=="browserPet")
end

function rematch:PetOnLeave()
	rematch:HideFloatingPetCard()
end

function rematch:PetOnClick(button)
	local cursorPetID = rematch:GetCursorPetID()
	if cursorPetID then -- if pet on the cursor
		if button=="RightButton" then
			RematchTooltip:Hide()
			ClearCursor()
			return
		end
		local receive = self:GetScript("OnReceiveDrag")
		if receive then -- any slot that expects to receive pets, run its function and leave
			return receive(self)
		end
	end
	local petID = rematch:GetPetID(self.petID)
	-- send to chat if Shift+Clicked (or whatever CHATLINK modifier they use)
	if IsModifiedClick("CHATLINK") and type(petID)=="string" then
		return ChatEdit_InsertLink(C_PetJournal.GetBattlePetLink(petID))
	end
	if button=="RightButton" then
		if self.menu then
			rematch:SetMenuSubject(petID,self:GetID())
			rematch:ShowMenu(self.menu,"cursor")
		end
	elseif petID then
		local cardID = RematchFloatingPetCard.petID
		rematch:LockFloatingPetCard()
		if petID ~= cardID then
			rematch:ShowFloatingPetCard(petID,self)
			rematch:LockFloatingPetCard() -- lock it again
		end
	end
end

function rematch:PetOnDoubleClick()
	local petID = rematch:GetPetID(self.petID)
	if type(petID)=="string" then
		C_PetJournal.SummonPetByGUID(petID)
	end
	rematch:HideFloatingPetCard(true)
end

function rematch:PetOnDragStart()
	if self.petID then
		C_PetJournal.PickupPet(self.petID)
	end
end

function rematch:AbilityOnEnter()
	local petID = self:GetParent().petID
	if petID and self.abilityID then
		rematch.ShowAbilityCard(self,petID,self.abilityID)
	end
end

function rematch:AbilityOnClick()
	if IsModifiedClick("CHATLINK") and self.abilityID and self:GetParent().petID then
		rematch:ChatLinkAbility(self:GetParent().petID,self.abilityID)
	end
end

--[[ ListTemplate script handlers ]]

function rematch:ListScrollToTop(frame)
	(frame or self:GetParent()).scrollBar:SetValue(0)
	PlaySound("UChatScrollButton")
end

function rematch:ListScrollToBottom(frame)
	local scrollFrame = frame or self:GetParent()
	scrollFrame.scrollBar:SetValue(scrollFrame.range)
	PlaySound("UChatScrollButton")
end

function rematch:ListScrollToIndex(frame,index)
	if index then
		local scrollFrame = frame or self:GetParent()
		if scrollFrame.scrollBar:IsEnabled() then
			local buttons = scrollFrame.buttons
			local height = math.max(0,floor(scrollFrame.buttonHeight*(index-((scrollFrame:GetHeight()/scrollFrame.buttonHeight))/2)))
			HybridScrollFrame_SetOffset(scrollFrame,height)
			scrollFrame.scrollBar:SetValue(height)
		else
			rematch:ListScrollToTop(frame)
		end
	end
end

-- used in scrollBar's OnValueChanged hooked in during the OnLoad
-- when scrolling with the mousewheel, force an OnEnter of whatever is under mouse
function rematch:ReOnEnter()
	local focus = GetMouseFocus()
	local scrollFrame = self:GetParent()
	for i=1,#scrollFrame.buttons do
		if scrollFrame.buttons[i]==focus then
			local script = focus:GetScript("OnEnter")
			if script then
				script(focus) -- if there's an OnEnter script, use it for focus
			end
			return
		end
	end
end

function rematch:ListOnLoad()
	-- tweak scrollBar appearance
	local scrollBar = self.scrollFrame.scrollBar
	scrollBar.doNotHide = true
	scrollBar.trackBG:SetPoint("TOPLEFT",-2,25)
	scrollBar.trackBG:SetPoint("BOTTOMRIGHT",2,-25)
	scrollBar.trackBG:SetTexture("Interface\\ChatFrame\\ChatFrameBackground")
	scrollBar.trackBG:SetGradientAlpha("HORIZONTAL",.125,.125,.125,1,.05,.05,.05,1)
	scrollBar:HookScript("OnValueChanged",rematch.ReOnEnter)
end

function rematch:ListOnSizeChanged()
	self.scrollFrame.stepSize = floor(self:GetHeight()-9.5) - self.scrollFrame.buttonHeight
end

--[[ Other template script handlers ]]

-- when a button is resized or released from being pushed, reset its icon to normal
function rematch:ResizeSlotIcon()
	local inset = ceil(self:GetWidth()/12)
	self.icon:SetPoint("TOPLEFT",inset,-inset)
	self.icon:SetPoint("BOTTOMRIGHT",-inset-1,inset+1)
	self.icon:SetVertexColor(1,1,1)
end

function rematch:PushToolbarButton()
	self.icon:SetSize(18,18)
	local vertex = self.vertex and self.vertex*0.65 or 0.65
	self.icon:SetVertexColor(vertex,vertex,vertex)
end

function rematch:ReleaseToolbarButton()
	self.icon:SetSize(20,20)
	local vertex = self.vertex or 1
	self.icon:SetVertexColor(vertex,vertex,vertex)
end

function rematch:ToolbarButtonOnEnter()
	if self.tooltipTitle then
		rematch:ShowTooltip(self.tooltipTitle,self.tooltipBody)
		RematchTooltip:ClearAllPoints()
		RematchTooltip:SetPoint("BOTTOM",self,"TOP",0,-2)
	end
end

function rematch:ScrollFrameToTop(scrollFrame)
	scrollFrame.scrollBar:SetValue(0)
end

function rematch:ScrollFrameToBottom(scrollFrame)
	scrollFrame.scrollBar:SetValue(scrollFrame.range)
end

--[[ Pet frames ]]

function rematch:WipePetFrames(pets)
	for i=1,3 do
		pets[i].petID = nil
		pets[i].icon:Hide()
		pets[i].dead:Hide()
		pets[i].leveling:Hide()
		pets[i].leveling:SetDesaturated(false)
		pets[i].leveling:SetVertexColor(1,1,1)
		for j=1,3 do
			pets[i].abilities[j].abilityID = nil
			pets[i].abilities[j].icon:Hide()
			if pets[i].abilities[j].level then
				pets[i].abilities[j].level:Hide()
			end
		end
	end
end

--[[ Timer Management ]]

rematch.timerFuncs = {} -- indexed by arbitrary name, the func to run when timer runs out
rematch.timerTimes = {} -- indexed by arbitrary name, the duration to run the timer
rematch.timerFrame = CreateFrame("Frame") -- timer independent of main frame visibility
rematch.timerFrame:Hide()
rematch.timersRunning = {} -- indexed numerically, timers that are running

function rematch:StartTimer(name,duration,func)
	local timers = rematch.timersRunning
	rematch.timerFuncs[name] = func
	rematch.timerTimes[name] = duration
	if not tContains(timers,name) then
		tinsert(timers,name)
	end
	rematch.timerFrame:Show()
end

function rematch:StopTimer(name)
	local timers = rematch.timersRunning
	for i=#timers,1,-1 do
		if timers[i]==name then
			tremove(timers,i)
			return
		end
	end
end

rematch.timerFrame:SetScript("OnUpdate",function(self,elapsed)
	local tick
	local times = rematch.timerTimes
	local timers = rematch.timersRunning

	for i=#timers,1,-1 do
		local name = timers[i]
		times[name] = times[name] - elapsed
		if times[name] < 0 then
			tremove(timers,i)
			rematch.timerFuncs[name]()
		end
		tick = true
	end

	if not tick then
		self:Hide()
	end
end)

--[[ Tooltips ]]

-- anchors frame to relativeTo depending on where relativeTo is on the screen, based on
-- the center of the reference frame (rematch frame itself if no reference given)
-- specifically, frame will be anchored to the corner furthest from the edge of the screen
function rematch:SmartAnchor(frame,relativeTo,reference)
	reference = reference or rematch
	local referenceScale = reference:GetEffectiveScale()
	local UIParentScale = UIParent:GetEffectiveScale()
	local isLeft = (reference:GetRight()*referenceScale+reference:GetLeft()*referenceScale)/2 < (UIParent:GetWidth()*UIParentScale)/2
	local isBottom = (reference:GetTop()*referenceScale+reference:GetBottom()*referenceScale)/2 < (UIParent:GetHeight()*UIParentScale)/2
	if isLeft then
		anchorPoint = isBottom and "BOTTOMLEFT" or "TOPLEFT"
		relativePoint = isBottom and "TOPRIGHT" or "BOTTOMRIGHT"
	else
		anchorPoint = isBottom and "BOTTOMRIGHT" or "TOPRIGHT"
		relativePoint = isBottom and "TOPLEFT" or "BOTTOMLEFT"
	end
	frame:ClearAllPoints()
	frame:SetPoint(anchorPoint,relativeTo,relativePoint)
end

-- shows tooltip with title[,body] and resizes around text
-- if priority is set, show through normal HideTooltip option and float by cursor with an OnUpdate
-- if no title is given, it will fetch self's .tooltipTitle and .tooltipBody
function rematch:ShowTooltip(title,body,priority)
	if not title then
		title = self.tooltipTitle
		body = self.tooltipBody
		priority = self.tooltipPriority
	end
	if title and not settings.HideTooltips or (priority and not settings.HideWarnings) then
		local tooltip = RematchTooltip
		tooltip.title:SetText(title)
		tooltip.body:SetText(body)
		local cx1=tooltip.title:GetStringWidth()
		local cx2=tooltip.body:GetStringWidth()
		local cy1=tooltip.title:GetStringHeight()
		local cy2=tooltip.body:GetStringHeight()
		tooltip:SetWidth( (cx1>cx2 and cx1 or cx2)+16 )
		tooltip:SetHeight( cy1+cy2+16 + ((body and body:len()>0 and 4) or 0) )
		tooltip:Show()
		if not priority then -- regular tooltips are SmartAnchored
			rematch:SmartAnchor(tooltip,GetMouseFocus())
			tooltip:SetScript("OnUpdate",nil)
		else -- priority tooltips get an OnUpdated anchor by the cursor
			tooltip:SetScript("OnUpdate",rematch.TooltipOnUpdate)
		end
	end
end

function rematch:TooltipOnUpdate(elapsed)
	local x,y = GetCursorPosition()
	local scale = UIParent:GetEffectiveScale()
	x = x/scale
	y = y/scale
	self:ClearAllPoints()
	if x>UIParent:GetWidth()/2 then -- right half of screen
		self:SetPoint("BOTTOMRIGHT",UIParent,"BOTTOMLEFT",x-2,y+2)
	else -- left half of screen
		self:SetPoint("BOTTOMLEFT",UIParent,"BOTTOMLEFT",x+2,y+2)
	end
end

--[[ Keep Summoned ]]

-- primarilyy for KeepSummoned, call this before pets are about to change
function rematch:SummonedPetMayChange()
	if settings.KeepSummoned then
		rematch.preLoadCompanion = rematch.preLoadCompanion or C_PetJournal.GetSummonedPetGUID()
		rematch:RegisterEvent("UNIT_PET")
		rematch:StartTimer("RestoreTimeout",1,rematch.RestoreTimeout)
	end
end

function rematch.events.UNIT_PET()
	rematch:StopTimer("RestoreTimeout")
	rematch:UnregisterEvent("UNIT_PET")
	rematch:StartTimer("RestoreCompanion",1.6,rematch.RestoreCompanion) -- wait a GCD before restoring
end

function rematch:RestoreCompanion()
	if not InCombatLockdown() then -- can't SummonPetByGUID in combat :(
		local nowSummoned = C_PetJournal.GetSummonedPetGUID()
		if not rematch.preLoadCompanion and nowSummoned then
			C_PetJournal.SummonPetByGUID(nowSummoned) -- something summoned, had nothing before
		elseif nowSummoned ~= rematch.preLoadCompanion then
			C_PetJournal.SummonPetByGUID(rematch.preLoadCompanion) -- something summoned different than before
		end
		rematch.preLoadCompanion = nil
	end
end

-- if no UNIT_PET fired, unregister it. apparently no pet was loaded
function rematch:RestoreTimeout()
	rematch.preLoadCompanion = nil
	rematch:UnregisterEvent("UNIT_PET")
end

--[[ Miscellaneous ]]

function rematch:GetCursorPetID()
	local cursorType,petID = GetCursorInfo()
	if cursorType=="battlepet" then
		return petID
	end
end

-- use this instead of a direct C_PetJournal.SetPetLoadOutInfo
function rematch:LoadPetSlot(slot,petID)
	if slot>0 and slot<4 and type(petID)=="string" then -- hilarious crash bug if we attempt to load slots other than 1-3
		rematch:SummonedPetMayChange()
		C_PetJournal.SetPetLoadOutInfo(slot,petID)
		if PetJournal then
			PetJournal_UpdatePetLoadOut() -- update petjournal if it's open
		end
	end
end

-- this makes the current pets glow when a pet is picked up onto the cursor
function rematch.events.CURSOR_UPDATE()
  local hasPet,petID = GetCursorInfo()
	hasPet = hasPet=="battlepet"
	for i=1,3 do
		rematch.current.pets[i].glow:SetShown(hasPet)
	end
	rematch.drawer.queue.levelingSlot.glow:SetShown(hasPet and rematch:PetCanLevel(petID))
	if hasPet then
		rematch:HideDialogs()
		rematch:RegisterEvent("CURSOR_UPDATE") -- this is the only place this event is registered
	else
		rematch:UnregisterEvent("CURSOR_UPDATE") -- cursor clear, stop watching cursor changes
	end
	if rematch.drawer.queue:IsVisible() then
		rematch:UpdateQueueList()
	end
end

-- for general purpose use, creates a shine cooldown effect on corner of frame
function rematch:ShineOnYouCrazy(frame,corner,xoff,yoff)
	if not rematch.shine then
		rematch.shine = CreateFrame("Cooldown")
		rematch.shine:SetSize(32,32)
	end
	rematch.shine:SetParent(frame:GetObjectType()=="Frame" and frame or frame:GetParent())
	rematch.shine:SetPoint("CENTER",frame,corner,xoff,yoff)
	rematch.shine:SetCooldown(0,1)
end

-- toggle tooltip
function rematch:ToggleTooltip()
	GameTooltip:SetOwner(self, self==RematchJournalButton and "ANCHOR_RIGHT" or "ANCHOR_LEFT")
	GameTooltip:SetText("Rematch",1,1,1)
	GameTooltip:AddLine(L["Toggles the Rematch window to manage battle pets and teams."],nil,nil,nil,true)
	GameTooltip:Show()
end
function rematch:ToggleTooltipHide()
	GameTooltip:Hide()
end

function rematch:GetPetID(petID)
	if petID==0 then
		return rematch:GetCurrentLevelingPet()
	else
		return petID
	end
end

function rematch:GetPetIDIcon(petID)
	local icon
	if petID==0 then
		return rematch.levelingIcon
	elseif not petID or petID==rematch.emptyPetID then
		return nil
	elseif type(petID)=="string" then
		return (select(9,C_PetJournal.GetPetInfoByPetID(petID)))
	else
		return (select(2,C_PetJournal.GetPetInfoBySpeciesID(petID))),true
	end
end

function rematch:ChatLinkAbility(petID,abilityID)
	local maxHealth,power,speed,_ = 100,0,0
	if type(petID)=="string" and petID~=rematch.emptyPetID then
		_,maxHealth,power,speed = C_PetJournal.GetPetStats(petID)
	end
	ChatEdit_InsertLink(GetBattlePetAbilityHyperlink(abilityID,maxHealth,power,speed))
end
